@@ -4,93 +4,36 @@
 use std::collections::{BTreeMap, HashMap};
 use std::fs::{
     create_dir, hard_link, remove_file, set_permissions, symlink_metadata, File, OpenOptions,
-    Permissions,
 };
 use std::io::prelude::*;
 use std::io::Error;
 use std::io::ErrorKind;
 use std::io::Result;
 use std::io::SeekFrom;
-use std::os::unix::fs::{chown, fchown, lchown, symlink, PermissionsExt};
+use std::os::unix::fs::{chown, fchown, lchown, symlink};
+use std::path::{Path, PathBuf};
 use std::process::ChildStdout;
 use std::process::Command;
 use std::process::Stdio;
 use std::time::SystemTime;
 
-use crate::libc::{set_modified, strftime_local};
+use crate::header::*;
+use crate::libc::{mknod, set_modified, strftime_local};
+use crate::seek_forward::SeekForward;
 
+mod header;
 mod libc;
-
-const CPIO_HEADER_LENGTH: u32 = 110;
-const CPIO_MAGIC_NUMBER: [u8; 6] = *b"070701";
-const PIPE_SIZE: usize = 65536;
-
-const MODE_PERMISSION_MASK: u32 = 0o007_777;
-const MODE_FILETYPE_MASK: u32 = 0o770_000;
-const FILETYPE_FIFO: u32 = 0o010_000;
-const FILETYPE_CHARACTER_DEVICE: u32 = 0o020_000;
-const FILETYPE_DIRECTORY: u32 = 0o040_000;
-const FILETYPE_BLOCK_DEVICE: u32 = 0o060_000;
-const FILETYPE_REGULAR_FILE: u32 = 0o100_000;
-const FILETYPE_SYMLINK: u32 = 0o120_000;
-const FILETYPE_SOCKET: u32 = 0o140_000;
+mod seek_forward;
 
 pub const LOG_LEVEL_WARNING: u32 = 5;
 pub const LOG_LEVEL_INFO: u32 = 7;
 pub const LOG_LEVEL_DEBUG: u32 = 8;
 
-pub trait SeekForward {
-    /// Seek forward to an offset, in bytes, in a stream.
-    ///
-    /// A seek beyond the end of a stream is allowed, but behavior is defined
-    /// by the implementation.
-    ///
-    /// # Errors
-    ///
-    /// Seeking can fail, for example because it might involve flushing a buffer.
-    fn seek_forward(&mut self, offset: u64) -> Result<()>;
-}
-
-impl SeekForward for File {
-    fn seek_forward(&mut self, offset: u64) -> Result<()> {
-        self.seek(SeekFrom::Current(offset.try_into().unwrap()))?;
-        Ok(())
-    }
-}
-
-impl SeekForward for ChildStdout {
-    fn seek_forward(&mut self, offset: u64) -> Result<()> {
-        let mut seek_reader = self.take(offset);
-        let mut remaining: usize = offset.try_into().unwrap();
-        let mut buffer = [0; PIPE_SIZE];
-        while remaining > 0 {
-            let read = seek_reader.read(&mut buffer)?;
-            remaining -= read;
-        }
-        Ok(())
-    }
-}
-
-impl SeekForward for &[u8] {
-    fn seek_forward(&mut self, offset: u64) -> Result<()> {
-        let mut seek_reader = std::io::Read::take(self, offset);
-        let mut buffer = Vec::new();
-        let read = seek_reader.read_to_end(&mut buffer)?;
-        if read < offset.try_into().unwrap() {
-            return Err(Error::new(
-                ErrorKind::UnexpectedEof,
-                format!("read only {} bytes, but {} wanted", read, offset),
-            ));
-        }
-        Ok(())
-    }
-}
-
 struct CpioFilenameReader<'a, R: Read + SeekForward> {
     file: &'a mut R,
 }
 
-impl<'a, R: Read + SeekForward> Iterator for CpioFilenameReader<'a, R> {
+impl<R: Read + SeekForward> Iterator for CpioFilenameReader<'_, R> {
     type Item = Result<String>;
 
     fn next(&mut self) -> Option<Self::Item> {
@@ -107,102 +50,6 @@ impl<'a, R: Read + SeekForward> Iterator for CpioFilenameReader<'a, R> {
     }
 }
 
-#[derive(Debug, PartialEq)]
-struct Header {
-    ino: u32,
-    mode: u32,
-    uid: u32,
-    gid: u32,
-    nlink: u32,
-    mtime: u32,
-    filesize: u32,
-    major: u32,
-    minor: u32,
-    // unused
-    //rmajor: u32,
-    //rminor: u32,
-    filename: String,
-}
-
-impl Header {
-    // Return major and minor combined as u64
-    fn dev(&self) -> u64 {
-        u64::from(self.major) << 32 | u64::from(self.minor)
-    }
-
-    fn mode_perm(&self) -> u32 {
-        self.mode & MODE_PERMISSION_MASK
-    }
-
-    // ls-style ASCII representation of the mode
-    fn mode_string(&self) -> [u8; 10] {
-        [
-            match self.mode & MODE_FILETYPE_MASK {
-                FILETYPE_FIFO => b'p',
-                FILETYPE_CHARACTER_DEVICE => b'c',
-                FILETYPE_DIRECTORY => b'd',
-                FILETYPE_BLOCK_DEVICE => b'b',
-                FILETYPE_REGULAR_FILE => b'-',
-                FILETYPE_SYMLINK => b'l',
-                FILETYPE_SOCKET => b's',
-                _ => b'?',
-            },
-            if self.mode & 0o400 != 0 { b'r' } else { b'-' },
-            if self.mode & 0o200 != 0 { b'w' } else { b'-' },
-            match self.mode & 0o4100 {
-                0o4100 => b's', // set-uid and executable by owner
-                0o4000 => b'S', // set-uid but not executable by owner
-                0o0100 => b'x',
-                _ => b'-',
-            },
-            if self.mode & 0o040 != 0 { b'r' } else { b'-' },
-            if self.mode & 0o020 != 0 { b'w' } else { b'-' },
-            match self.mode & 0o2010 {
-                0o2010 => b's', // set-gid and executable by group
-                0o2000 => b'S', // set-gid but not executable by group
-                0o0010 => b'x',
-                _ => b'-',
-            },
-            if self.mode & 0o004 != 0 { b'r' } else { b'-' },
-            if self.mode & 0o002 != 0 { b'w' } else { b'-' },
-            match self.mode & 0o1001 {
-                0o1001 => b't', // sticky and executable by others
-                0o1000 => b'T', // sticky but not executable by others
-                0o0001 => b'x',
-                _ => b'-',
-            },
-        ]
-    }
-
-    fn permission(&self) -> Permissions {
-        PermissionsExt::from_mode(self.mode & MODE_PERMISSION_MASK)
-    }
-
-    fn ino_and_dev(&self) -> u128 {
-        u128::from(self.ino) << 64 | u128::from(self.dev())
-    }
-
-    fn mark_seen(&self, seen_files: &mut SeenFiles) {
-        seen_files.insert(self.ino_and_dev(), self.filename.clone());
-    }
-
-    fn read_symlink_target<R: Read>(&self, file: &mut R) -> Result<String> {
-        let align = align_to_4_bytes(self.filesize);
-        let mut target_bytes = vec![0u8; (self.filesize + align).try_into().unwrap()];
-        file.read_exact(&mut target_bytes)?;
-        target_bytes.truncate(self.filesize.try_into().unwrap());
-        // TODO: propper name reading handling
-        let target = std::str::from_utf8(&target_bytes).unwrap();
-        Ok(target.into())
-    }
-
-    fn skip_file_content<R: SeekForward>(&self, file: &mut R) -> Result<()> {
-        let skip = self.filesize + align_to_4_bytes(self.filesize);
-        file.seek_forward(skip.into())?;
-        Ok(())
-    }
-}
-
 struct UserGroupCache {
     user_cache: HashMap<u32, Option<String>>,
     group_cache: HashMap<u32, Option<String>>,
@@ -291,93 +138,12 @@ fn align_to_4_bytes(length: u32) -> u32 {
     }
 }
 
-fn hex_str_to_u32(bytes: &[u8]) -> Result<u32> {
-    let s = match std::str::from_utf8(bytes) {
-        Err(_) => {
-            return Err(Error::new(
-                ErrorKind::InvalidData,
-                format!("Invalid hexadecimal value '{}'", bytes.escape_ascii()),
-            ))
-        }
-        Ok(value) => value,
-    };
-    match u32::from_str_radix(s, 16) {
-        Err(_) => Err(Error::new(
-            ErrorKind::InvalidData,
-            format!("Invalid hexadecimal value '{}'", s),
-        )),
-        Ok(value) => Ok(value),
-    }
-}
-
-fn read_filename<R: Read>(file: &mut R, namesize: u32) -> Result<String> {
-    let header_align = align_to_4_bytes(CPIO_HEADER_LENGTH + namesize);
-    let mut filename_bytes = vec![0u8; (namesize + header_align).try_into().unwrap()];
-    let filename_length: usize = (namesize - 1).try_into().unwrap();
-    file.read_exact(&mut filename_bytes)?;
-    if filename_bytes[filename_length] != 0 {
-        return Err(Error::new(
-            ErrorKind::InvalidData,
-            format!(
-                "Entry name '{:?}' is not NULL-terminated",
-                &filename_bytes[0..filename_length]
-            ),
-        ));
-    }
-    filename_bytes.truncate(filename_length);
-    // TODO: propper name reading handling
-    let filename = std::str::from_utf8(&filename_bytes).unwrap();
-    Ok(filename.to_string())
-}
-
-fn check_begins_with_cpio_magic_header(header: &[u8]) -> std::io::Result<()> {
-    if header[0..6] != CPIO_MAGIC_NUMBER {
-        return Err(Error::new(
-            ErrorKind::InvalidData,
-            format!(
-                "Invalid CPIO magic number '{}'. Expected {}",
-                &header[0..6].escape_ascii(),
-                std::str::from_utf8(&CPIO_MAGIC_NUMBER).unwrap(),
-            ),
-        ));
-    }
-    Ok(())
-}
-
-fn read_cpio_header<R: Read>(file: &mut R) -> std::io::Result<Header> {
-    let mut buffer = [0; CPIO_HEADER_LENGTH as usize];
-    file.read_exact(&mut buffer)?;
-    check_begins_with_cpio_magic_header(&buffer)?;
-    let namesize = hex_str_to_u32(&buffer[94..102])?;
-    let filename = read_filename(file, namesize)?;
-    Ok(Header {
-        ino: hex_str_to_u32(&buffer[6..14])?,
-        mode: hex_str_to_u32(&buffer[14..22])?,
-        uid: hex_str_to_u32(&buffer[22..30])?,
-        gid: hex_str_to_u32(&buffer[30..38])?,
-        nlink: hex_str_to_u32(&buffer[38..46])?,
-        mtime: hex_str_to_u32(&buffer[46..54])?,
-        filesize: hex_str_to_u32(&buffer[54..62])?,
-        major: hex_str_to_u32(&buffer[62..70])?,
-        minor: hex_str_to_u32(&buffer[70..78])?,
-        //rmajor: hex_str_to_u32(&buffer[78..86])?,
-        //rminor: hex_str_to_u32(&buffer[86..94])?,
-        filename,
-    })
-}
-
 /// Read only the file name from the next cpio object.
 ///
 /// Read the next cpio object header, check the magic, skip the file data.
 /// Return the file name.
 fn read_filename_from_next_cpio_object<R: Read + SeekForward>(file: &mut R) -> Result<String> {
-    let mut header = [0; CPIO_HEADER_LENGTH as usize];
-    file.read_exact(&mut header)?;
-    check_begins_with_cpio_magic_header(&header)?;
-    let filesize = hex_str_to_u32(&header[54..62])?;
-    let namesize = hex_str_to_u32(&header[94..102])?;
-    let filename = read_filename(file, namesize)?;
-
+    let (filesize, filename) = Header::read_only_filesize_and_filename(file)?;
     let skip = filesize + align_to_4_bytes(filesize);
     file.seek_forward(skip.into())?;
     Ok(filename)
@@ -496,7 +262,7 @@ fn read_cpio_and_print_long_format<R: Read + SeekForward, W: Write>(
     let mut last_mtime = 0;
     let mut time_string: String = "".into();
     loop {
-        let header = match read_cpio_header(file) {
+        let header = match Header::read(file) {
             Ok(header) => {
                 if header.filename == "TRAILER!!!" {
                     break;
@@ -521,33 +287,51 @@ fn read_cpio_and_print_long_format<R: Read + SeekForward, W: Write>(
             time_string = format_time(header.mtime, now)?;
         };
 
-        if header.mode & MODE_FILETYPE_MASK == FILETYPE_SYMLINK {
-            let target = header.read_symlink_target(file)?;
-            writeln!(
-                out,
-                "{} {:>3} {:<8} {:<8} {:>8} {} {} -> {}",
-                std::str::from_utf8(&mode_string).unwrap(),
-                header.nlink,
-                user,
-                group,
-                header.filesize,
-                time_string,
-                header.filename,
-                target
-            )?;
-        } else {
-            header.skip_file_content(file)?;
-            writeln!(
-                out,
-                "{} {:>3} {:<8} {:<8} {:>8} {} {}",
-                std::str::from_utf8(&mode_string).unwrap(),
-                header.nlink,
-                user,
-                group,
-                header.filesize,
-                time_string,
-                header.filename
-            )?;
+        match header.mode & MODE_FILETYPE_MASK {
+            FILETYPE_SYMLINK => {
+                let target = header.read_symlink_target(file)?;
+                writeln!(
+                    out,
+                    "{} {:>3} {:<8} {:<8} {:>8} {} {} -> {}",
+                    std::str::from_utf8(&mode_string).unwrap(),
+                    header.nlink,
+                    user,
+                    group,
+                    header.filesize,
+                    time_string,
+                    header.filename,
+                    target
+                )?;
+            }
+            FILETYPE_BLOCK_DEVICE | FILETYPE_CHARACTER_DEVICE => {
+                header.skip_file_content(file)?;
+                writeln!(
+                    out,
+                    "{} {:>3} {:<8} {:<8} {:>3}, {:>3} {} {}",
+                    std::str::from_utf8(&mode_string).unwrap(),
+                    header.nlink,
+                    user,
+                    group,
+                    header.rmajor,
+                    header.rminor,
+                    time_string,
+                    header.filename
+                )?;
+            }
+            _ => {
+                header.skip_file_content(file)?;
+                writeln!(
+                    out,
+                    "{} {:>3} {:<8} {:<8} {:>8} {} {}",
+                    std::str::from_utf8(&mode_string).unwrap(),
+                    header.nlink,
+                    user,
+                    group,
+                    header.filesize,
+                    time_string,
+                    header.filename
+                )?;
+            }
         };
     }
     Ok(())
@@ -567,6 +351,47 @@ fn create_dir_ignore_existing<P: AsRef<std::path::Path>>(path: P) -> Result<()>
     Ok(())
 }
 
+fn write_character_device(
+    header: &Header,
+    preserve_permissions: bool,
+    log_level: u32,
+) -> Result<()> {
+    if header.filesize != 0 {
+        return Err(Error::new(
+            ErrorKind::InvalidData,
+            format!(
+                "Invalid size for character device '{}': {} bytes instead of 0.",
+                header.filename, header.filesize
+            ),
+        ));
+    };
+    if log_level >= LOG_LEVEL_DEBUG {
+        writeln!(
+            std::io::stderr(),
+            "Creating character device '{}' with mode {:o}",
+            header.filename,
+            header.mode_perm(),
+        )?;
+    };
+    if let Err(e) = mknod(&header.filename, header.mode, header.rmajor, header.rminor) {
+        match e.kind() {
+            ErrorKind::AlreadyExists => {
+                remove_file(&header.filename)?;
+                mknod(&header.filename, header.mode, header.rmajor, header.rminor)?;
+            }
+            _ => {
+                return Err(e);
+            }
+        }
+    };
+    if preserve_permissions {
+        lchown(&header.filename, Some(header.uid), Some(header.gid))?;
+    };
+    set_permissions(&header.filename, header.permission())?;
+    set_modified(&header.filename, header.mtime.into())?;
+    Ok(())
+}
+
 fn write_directory(
     header: &Header,
     preserve_permissions: bool,
@@ -608,13 +433,6 @@ fn from_mtime(mtime: u32) -> SystemTime {
     std::time::UNIX_EPOCH + std::time::Duration::from_secs(mtime.into())
 }
 
-fn try_get_hard_link_target<'a>(header: &Header, seen_files: &'a SeenFiles) -> Option<&'a String> {
-    if header.nlink <= 1 {
-        return None;
-    }
-    seen_files.get(&header.ino_and_dev())
-}
-
 fn write_file<R: Read + SeekForward>(
     cpio_file: &mut R,
     header: &Header,
@@ -623,7 +441,7 @@ fn write_file<R: Read + SeekForward>(
     log_level: u32,
 ) -> Result<()> {
     let mut file;
-    if let Some(target) = try_get_hard_link_target(header, seen_files) {
+    if let Some(target) = header.try_get_hard_link_target(seen_files) {
         if log_level >= LOG_LEVEL_DEBUG {
             writeln!(
                 std::io::stderr(),
@@ -733,14 +551,53 @@ fn write_symbolic_link<R: Read + SeekForward>(
     Ok(())
 }
 
+fn absolute_parent_directory<S: AsRef<str>>(path: S, base_dir: &Path) -> Result<PathBuf>
+where
+    PathBuf: From<S>,
+{
+    let abspath = if path.as_ref().starts_with("/") {
+        PathBuf::from(path)
+    } else {
+        base_dir.join(path.as_ref())
+    };
+    match abspath.parent() {
+        Some(d) => Ok(d.into()),
+        // TODO: Use ErrorKind::InvalidFilename once stable.
+        None => Err(Error::new(
+            ErrorKind::InvalidData,
+            format!("Path {:#?} has no parent directory.", abspath),
+        )),
+    }
+}
+
+fn check_path_is_canonical_subdir<S: AsRef<str> + std::fmt::Display>(
+    path: S,
+    dir: &Path,
+    base_dir: &PathBuf,
+) -> Result<PathBuf> {
+    let canonicalized_path = dir.canonicalize()?;
+    if !canonicalized_path.starts_with(base_dir) {
+        return Err(Error::new(
+            ErrorKind::InvalidData,
+            format!(
+                "The parent directory of \"{}\" (resolved to {:#?}) is not within the directory {:#?}.",
+                path, canonicalized_path, base_dir
+            ),
+        ));
+    }
+    Ok(canonicalized_path)
+}
+
 fn read_cpio_and_extract<R: Read + SeekForward>(
     file: &mut R,
+    base_dir: &PathBuf,
     preserve_permissions: bool,
     log_level: u32,
 ) -> Result<()> {
     let mut extractor = Extractor::new();
+    let mut previous_checked_dir = PathBuf::new();
     loop {
-        let header = match read_cpio_header(file) {
+        let header = match Header::read(file) {
             Ok(header) => {
                 if header.filename == "TRAILER!!!" {
                     break;
@@ -757,7 +614,21 @@ fn read_cpio_and_extract<R: Read + SeekForward>(
             writeln!(std::io::stderr(), "{}", header.filename)?;
         }
 
+        if !header.is_root_directory() {
+            let absdir = absolute_parent_directory(&header.filename, base_dir)?;
+            // canonicalize() is an expensive call. So cache the previously resolved
+            // parent directory. Skip the path traversal check in case the absolute
+            // parent directory has no symlinks and matches the previouly checked directory.
+            if absdir != previous_checked_dir {
+                previous_checked_dir =
+                    check_path_is_canonical_subdir(&header.filename, &absdir, base_dir)?;
+            }
+        }
+
         match header.mode & MODE_FILETYPE_MASK {
+            FILETYPE_CHARACTER_DEVICE => {
+                write_character_device(&header, preserve_permissions, log_level)?
+            }
             FILETYPE_DIRECTORY => write_directory(
                 &header,
                 preserve_permissions,
@@ -774,7 +645,7 @@ fn read_cpio_and_extract<R: Read + SeekForward>(
             FILETYPE_SYMLINK => {
                 write_symbolic_link(file, &header, preserve_permissions, log_level)?
             }
-            FILETYPE_FIFO | FILETYPE_CHARACTER_DEVICE | FILETYPE_BLOCK_DEVICE | FILETYPE_SOCKET => {
+            FILETYPE_FIFO | FILETYPE_BLOCK_DEVICE | FILETYPE_SOCKET => {
                 unimplemented!(
                     "Mode {:o} (file {}) not implemented. Please open a bug report requesting support for this type.",
                     header.mode, header.filename
@@ -803,6 +674,29 @@ fn seek_to_cpio_end(file: &mut File) -> Result<()> {
     Ok(())
 }
 
+pub fn get_cpio_archive_count(file: &mut File) -> Result<u32> {
+    let mut count = 0;
+    loop {
+        let command = match read_magic_header(file) {
+            None => return Ok(count),
+            Some(x) => x?,
+        };
+        count += 1;
+        if command.get_program() == "cpio" {
+            seek_to_cpio_end(file)?;
+        } else {
+            break;
+        }
+    }
+    Ok(count)
+}
+
+pub fn print_cpio_archive_count<W: Write>(mut file: File, out: &mut W) -> Result<()> {
+    let count = get_cpio_archive_count(&mut file)?;
+    writeln!(out, "{}", count)?;
+    Ok(())
+}
+
 pub fn examine_cpio_content<W: Write>(mut file: File, out: &mut W) -> Result<()> {
     loop {
         let command = match read_magic_header(&mut file) {
@@ -844,10 +738,15 @@ pub fn extract_cpio_archive(
             Some(x) => x?,
         };
         if command.get_program() == "cpio" {
-            read_cpio_and_extract(&mut file, preserve_permissions, log_level)?;
+            read_cpio_and_extract(&mut file, &base_dir, preserve_permissions, log_level)?;
         } else {
             let mut decompressed = decompress(&mut command, file)?;
-            read_cpio_and_extract(&mut decompressed, preserve_permissions, log_level)?;
+            read_cpio_and_extract(
+                &mut decompressed,
+                &base_dir,
+                preserve_permissions,
+                log_level,
+            )?;
             break;
         }
         count += 1;
@@ -894,10 +793,11 @@ pub fn list_cpio_content<W: Write>(mut file: File, out: &mut W, log_level: u32)
 
 #[cfg(test)]
 mod tests {
-    use std::env;
+    use std::env::{self, current_dir, set_current_dir};
+    use std::os::unix::fs::{FileTypeExt, MetadataExt, PermissionsExt};
 
     use super::*;
-    use std::os::unix::fs::MetadataExt;
+    use crate::libc::{major, minor};
 
     fn getgid() -> u32 {
         unsafe { ::libc::getgid() }
@@ -911,6 +811,33 @@ mod tests {
         fn tzset();
     }
 
+    struct TempDir {
+        path: PathBuf,
+        cwd: PathBuf,
+    }
+
+    impl TempDir {
+        fn new() -> Result<Self> {
+            // Use some very pseudo-random number
+            let cwd = current_dir()?;
+            let epoch = SystemTime::now()
+                .duration_since(std::time::UNIX_EPOCH)
+                .unwrap();
+            let name = std::option_env!("CARGO_PKG_NAME").unwrap();
+            let dir_builder = std::fs::DirBuilder::new();
+            let mut path = env::temp_dir();
+            path.push(format!("{}-{}", name, epoch.subsec_nanos()));
+            dir_builder.create(&path).map(|_| Self { path, cwd })
+        }
+    }
+
+    impl Drop for TempDir {
+        fn drop(&mut self) {
+            let _ = set_current_dir(&self.cwd);
+            let _ = std::fs::remove_dir_all(&self.path);
+        }
+    }
+
     impl UserGroupCache {
         fn insert_test_data(&mut self) {
             self.user_cache.insert(1000, Some("user".into()));
@@ -919,6 +846,45 @@ mod tests {
         }
     }
 
+    #[test]
+    fn test_absolute_parent_directory() {
+        let base_dir = Path::new("/nonexistent/arthur");
+        assert_eq!(
+            absolute_parent_directory("usr/bin/true", base_dir).unwrap(),
+            PathBuf::from("/nonexistent/arthur/usr/bin")
+        );
+        assert_eq!(
+            absolute_parent_directory("/usr/bin/true", base_dir).unwrap(),
+            PathBuf::from("/usr/bin")
+        );
+        assert_eq!(
+            absolute_parent_directory(".", base_dir).unwrap(),
+            PathBuf::from("/nonexistent")
+        );
+    }
+
+    // Test detecting path traversal attacks like CVE-2015-1197
+    #[test]
+    fn test_read_cpio_and_extract_path_traversal() {
+        let mut file = File::open("tests/path-traversal.cpio").unwrap();
+        let tempdir = TempDir::new().unwrap();
+        set_current_dir(&tempdir.path).unwrap();
+        let got =
+            read_cpio_and_extract(&mut file, &tempdir.path, false, LOG_LEVEL_WARNING).unwrap_err();
+        assert_eq!(got.kind(), ErrorKind::InvalidData);
+        assert_eq!(got.to_string(), format!(
+            "The parent directory of \"tmp/trav.txt\" (resolved to \"/tmp\") is not within the directory {:#?}.",
+            &tempdir.path
+        ));
+    }
+
+    #[test]
+    fn test_absolute_parent_directory_error() {
+        let got = absolute_parent_directory(".", Path::new("/")).unwrap_err();
+        assert_eq!(got.kind(), ErrorKind::InvalidData);
+        assert_eq!(got.to_string(), "Path \"/.\" has no parent directory.");
+    }
+
     #[test]
     fn test_align_to_4_bytes() {
         assert_eq!(align_to_4_bytes(110), 2);
@@ -942,23 +908,49 @@ mod tests {
     }
 
     #[test]
-    fn test_hex_str_to_u32() {
-        let value = hex_str_to_u32(b"000003E8").unwrap();
-        assert_eq!(value, 1000);
+    fn test_get_cpio_archive_count_single() {
+        let mut file = File::open("tests/single.cpio").expect("test cpio should be present");
+        let count = get_cpio_archive_count(&mut file).unwrap();
+        assert_eq!(count, 1);
     }
 
     #[test]
-    fn test_hex_str_to_u32_invalid_hex() {
-        let got = hex_str_to_u32(b"something").unwrap_err();
-        assert_eq!(got.kind(), ErrorKind::InvalidData);
-        assert_eq!(got.to_string(), "Invalid hexadecimal value 'something'");
+    fn test_print_cpio_archive_count() {
+        let mut file = File::open("tests/zstd.cpio").expect("test cpio should be present");
+        let mut output = Vec::new();
+
+        let count = get_cpio_archive_count(&mut file).unwrap();
+        assert_eq!(count, 2);
+
+        file.seek(SeekFrom::Start(0)).unwrap();
+        print_cpio_archive_count(file, &mut output).unwrap();
+        assert_eq!(String::from_utf8(output).unwrap(), "2\n");
     }
 
     #[test]
-    fn test_hex_str_to_u32_invalid_utf8() {
-        let got = hex_str_to_u32(b"no\xc3\x28utf8").unwrap_err();
-        assert_eq!(got.kind(), ErrorKind::InvalidData);
-        assert_eq!(got.to_string(), "Invalid hexadecimal value 'no\\xc3(utf8'");
+    fn test_read_cpio_and_print_long_format_character_device() {
+        // Wrapped before mtime and filename
+        let cpio_data = b"07070100000003000021A4000000000000\
+        00000000000167055BC800000000000000000000000000000005000000010000\
+        000C00000000dev/console\0\0\0\
+        0707010000000000000000000000000000000000000001\
+        0000000000000000000000000000000000000000000000000000000B00000000\
+        TRAILER!!!\0\0\0\0";
+        let mut output = Vec::new();
+        let mut user_group_cache = UserGroupCache::new();
+        env::set_var("TZ", "UTC");
+        unsafe { tzset() };
+        read_cpio_and_print_long_format(
+            &mut cpio_data.as_ref(),
+            &mut output,
+            1728486311,
+            &mut user_group_cache,
+        )
+        .unwrap();
+        assert_eq!(
+            String::from_utf8(output).unwrap(),
+            "crw-r--r--   1 root     root       5,   1 Oct  8 16:20 dev/console\n"
+        );
     }
 
     #[test]
@@ -1042,56 +1034,41 @@ mod tests {
     }
 
     #[test]
-    fn test_read_cpio_header() {
-        // Wrapped before mtime and filename
-        let cpio_data = b"07070100000002000081B4000003E8000007D000000001\
-            661BE5C600000008000000000000000000000000000000000000000A00000000\
-            path/file\0content\0";
-        let header = read_cpio_header(&mut cpio_data.as_ref()).unwrap();
-        assert_eq!(
-            header,
-            Header {
-                ino: 2,
-                mode: 0o100664,
-                uid: 1000,
-                gid: 2000,
-                nlink: 1,
-                mtime: 1713104326,
-                filesize: 8,
-                major: 0,
-                minor: 0,
-                filename: "path/file".into()
-            }
-        )
-    }
+    fn test_write_character_device() {
+        if getuid() != 0 {
+            // This test needs to run as root.
+            return;
+        }
+        let mut header = Header::new(1, 0o20_644, 0, 0, 0, 1740402179, 0, "./null");
+        header.rmajor = 1;
+        header.rminor = 3;
+        write_character_device(&header, true, LOG_LEVEL_WARNING).unwrap();
 
-    #[test]
-    fn test_read_cpio_header_invalid_magic_number() {
-        let invalid_data = b"abc\tefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ\
-            abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";
-        let got = read_cpio_header(&mut invalid_data.as_ref()).unwrap_err();
-        assert_eq!(got.kind(), ErrorKind::InvalidData);
-        assert_eq!(
-            got.to_string(),
-            "Invalid CPIO magic number 'abc\\tef'. Expected 070701"
-        );
+        let attr = std::fs::metadata("null").unwrap();
+        assert_eq!(attr.len(), header.filesize.into());
+        assert!(attr.file_type().is_char_device());
+        assert_eq!(attr.modified().unwrap(), from_mtime(header.mtime));
+        assert_eq!(attr.permissions(), PermissionsExt::from_mode(header.mode));
+        assert_eq!(attr.uid(), header.uid);
+        assert_eq!(attr.gid(), header.gid);
+        assert_eq!(major(attr.rdev()), header.rmajor);
+        assert_eq!(minor(attr.rdev()), header.rminor);
+        std::fs::remove_file("null").unwrap();
     }
 
     #[test]
     fn test_write_directory_with_setuid() {
         let mut mtimes = BTreeMap::new();
-        let header = Header {
-            ino: 1,
-            mode: 0o43_777,
-            uid: getuid(),
-            gid: getgid(),
-            nlink: 0,
-            mtime: 1720081471,
-            filesize: 0,
-            major: 0,
-            minor: 0,
-            filename: "./directory_with_setuid".into(),
-        };
+        let header = Header::new(
+            1,
+            0o43_777,
+            getuid(),
+            getgid(),
+            0,
+            1720081471,
+            0,
+            "./directory_with_setuid",
+        );
         write_directory(&header, true, LOG_LEVEL_WARNING, &mut mtimes).unwrap();
 
         let attr = std::fs::metadata("directory_with_setuid").unwrap();
@@ -1109,18 +1086,16 @@ mod tests {
     #[test]
     fn test_write_file_with_setuid() {
         let mut seen_files = SeenFiles::new();
-        let header = Header {
-            ino: 1,
-            mode: 0o104_755,
-            uid: getuid(),
-            gid: getgid(),
-            nlink: 0,
-            mtime: 1720081471,
-            filesize: 9,
-            major: 0,
-            minor: 0,
-            filename: "./file_with_setuid".into(),
-        };
+        let header = Header::new(
+            1,
+            0o104_755,
+            getuid(),
+            getgid(),
+            0,
+            1720081471,
+            9,
+            "./file_with_setuid",
+        );
         let cpio = b"!/bin/sh\n\0\0\0";
         write_file(
             &mut cpio.as_ref(),
@@ -1143,18 +1118,16 @@ mod tests {
 
     #[test]
     fn test_write_symbolic_link() {
-        let header = Header {
-            ino: 1,
-            mode: 0o120_777,
-            uid: getuid(),
-            gid: getgid(),
-            nlink: 0,
-            mtime: 1721427072,
-            filesize: 12,
-            major: 0,
-            minor: 0,
-            filename: "./dead_symlink".into(),
-        };
+        let header = Header::new(
+            1,
+            0o120_777,
+            getuid(),
+            getgid(),
+            0,
+            1721427072,
+            12,
+            "./dead_symlink",
+        );
         let cpio = b"/nonexistent";
         write_symbolic_link(&mut cpio.as_ref(), &header, true, LOG_LEVEL_WARNING).unwrap();
 
